<?php

namespace InSegment\WorkersFramework\Models;

use Shavshukov\EventAppBase\Model\ModelEventsTrait;
use Shavshukov\EventAppBase\Model\ArrayModel;
use Shavshukov\EventAppBase\Interfaces\StorageInterface;
use Shavshukov\EventAppBase\MicroEvent;

use InSegment\WorkersFramework\Interfaces\JobListInterface;
use InSegment\WorkersFramework\Queues\ConnectionsConfig;
use InSegment\WorkersFramework\Queues\JobList;

class Config extends ArrayModel
{
    use ModelEventsTrait;
    
    const DEFAULT_CONFIG_FILE  = 'config.yaml';
    
    const MANDATORY_ARRAYS = ['connections', 'workers', 'jobs', 'multi_master'];
    const CONFIG_MANDATORY_FIELDS = ['connections', 'workers', 'jobs', 'node_name', 'main_queue', 'service_queue', 'worker_run_command'];
    const WORKER_MANDATORY_FIELDS = ['work_queue', 'service_queue'];
    const MULTI_MASTER_MANDATORY_FIELDS = ['service_queue'];
    
    /**
     * Connections config instance
     *
     * @var \InSegment\WorkersFramework\Queues\ConnectionsConfig|null
     */
    protected $connectionsConfig;
    
    /**
     * Job list instance
     *
     * @var \InSegment\WorkersFramework\Interfaces\JobListInterface|null
     */
    protected $jobList;
    
    /**
     * Constructor
     */
    public function __construct()
    {
        $this->bootEventsIfNotBooted();
    }
    
    /**
     * @inheritdoc
     */
    public static function extend(StorageInterface $storage)
    {
        // TODO: move this method to model events trait
        $model = parent::extend($storage);
        
        MicroEvent::fireEvent(static::class, 'loading', $model, true);
        MicroEvent::fireEvent(static::class, 'loaded', $model, true);
        
        return $model;
    }
    
    /**
     * Get connections configuration
     * 
     * @return \InSegment\WorkersFramework\Queues\ConnectionsConfig
     */
    public function getConnectionsConfig(): ConnectionsConfig
    {
        return $this->connectionsConfig;
    }
    
    /**
     * Get list of jobs
     * 
     * @return \InSegment\WorkersFramework\Interfaces\JobListInterface
     */
    public function getJobList(): JobListInterface
    {
        return $this->jobList;
    }
    
    /**
     * Register events once on the class
     * 
     * @return null
     */
    protected static function bootEvents()
    {
        static::loaded(function ($model, $result) {
            if (!$result) {
                throw new \Exception('Unable to load config file');
            }
            
            if (($notFoundFields = array_diff(self::CONFIG_MANDATORY_FIELDS, array_keys($model->data)))) {
                $printNotFoundFields = implode(', ', $notFoundFields);
                throw new \Exception("Config checking: Fields not found: {$printNotFoundFields}");
            }
            
            foreach (self::MANDATORY_ARRAYS as $arrayName) {
                if (isset($model->data[$arrayName]) && !is_array($model->data[$arrayName])) {
                    throw new \Exception("Config checking: '{$arrayName}' is not an array");
                }
            }
            
            $model->connectionsConfig = new ConnectionsConfig($model->data['connections']);
            $model->jobList = new JobList($model->data['jobs']);
            
            foreach ($model->data['workers'] as $workerKey => $workerConfig) {
                if (($notFoundFields = array_diff(self::WORKER_MANDATORY_FIELDS, array_keys($workerConfig)))) {
                    $printNotFoundFields = implode(', ', $notFoundFields);
                    throw new \Exception("Workers config checking: Fields not found: {$printNotFoundFields} for key {$workerKey}");
                }
            }
            
            if (isset($model->data['multi_master'])) {
                foreach ($model->data['multi_master'] as $nodeName => $nodeConfig) {
                    if (($notFoundFields = array_diff(self::MULTI_MASTER_MANDATORY_FIELDS, array_keys($nodeConfig)))) {
                        $printNotFoundFields = implode(', ', $notFoundFields);
                        throw new \Exception("Multi-master config checking: Fields not found: {$printNotFoundFields} for node {$nodeName}");
                    }
                }
            }
            
            return true;
        });
    }
}
