<?php

namespace InSegment\WorkersFramework\Services;

use InSegment\WorkersFramework\Interfaces\LockfileInterface;
use Shavshukov\EventAppBase\ApplicationInterface;
use React\Stream\WritableResourceStream;

class WorkerGc implements LockfileInterface
{
    /**
     * Do GC of the lock file every N stopped processes
     * 
     * @var int
     */
    const GC_INTERVAL = 100;

    /**
     * Application instance
     * 
     * @var \Shavshukov\EventAppBase\ApplicationInterface $app 
     */
    protected $app;

    /**
     * Full path to the workers lock file
     * 
     * @var string|null
     */
    protected ?string $lockfile;

    /**
     * Started processes
     * 
     * @var array [
     *      int $PID => \React\ChildProcess\Process $process,
     * ]
     */
    protected array $startedProcesses;

    /**
     * Number of the child processes closed
     * 
     * @var int
     */
    protected int $closedProcessesCounter = 0;

    /**
     * The stream to write to the lockfile.
     * 
     * @var WritableResourceStream
     */
    protected WritableResourceStream $lockfileWritableStream;

    /**
     * Constructor
     * 
     * @param \Shavshukov\EventAppBase\ApplicationInterface $app
     */
    public function __construct(ApplicationInterface $app)
    {
        $this->app = $app;
        $this->lockfile = $this->app->config('worker_pid_lockfile');

        if ($this->lockfile) {
            $this->lockfileWritableStream = new WritableResourceStream(fopen($this->lockfile, 'w'), $this->app->getLoop());
            $this->app->getLoop()->addPeriodicTimer(1, function () {
                $this->workerLockfileGc();
            });
            echo "Workers PID lockfile GC started\n";
        }
    }

    public function register(\React\ChildProcess\Process $process)
    {
        if (!$this->lockfile) {
            return;
        }

        $childPid = $process->getPid();
        echo "Spawning child process PID {$childPid}\n";
        if ($this->lockfileWritableStream->isWritable()) {
            $this->lockfileWritableStream->write("{$childPid}\n");
        } else {
            echo "Warning: the lockfile stream is not writable. Can not write PID: {$childPid}\n";
        }
        $this->startedProcesses[$childPid] = $process;
    }

    protected function workerLockfileGc()
    {
        foreach ($this->startedProcesses as $pid => $process) {
            if ($process->isStopped() || $process->isTerminated() || !$process->isRunning()) {
                unset($this->startedProcesses[$pid]);
                ++$this->closedProcessesCounter;
            }
        }

        if ($this->closedProcessesCounter !== self::GC_INTERVAL) {
            return;
        }
        $this->closedProcessesCounter = 0;

        echo "GC workers PID lockfile...\n";
        $handler = fopen($this->lockfile, 'r+');
        if (flock($handler, LOCK_EX)) {
            ftruncate($handler, 0);
            fwrite($handler, implode("\n", array_keys($this->startedProcesses)));
            fflush($handler);
            flock($handler, LOCK_UN);
        } else {
            throw new \Exception(__METHOD__ . ": can not lock file {$this->lockfile} for garbage collecting");
        }
        fclose($handler);

        echo "GC success\n";
    }
}
